"""Generated Tkinter Elements to Create a Custom Title Bar"""
from . import elements as E
from .elements import tk, ttk

_ = tk, ttk


class Type:
    def __class_getitem__(cls, item: str): return getattr(cls, item, False)

    def __init__(self, *_, **__: E.Any):
        super().__init__(*_, **__)


class Widget(Type):
    _ = Frame, Label, Button = E.Frame, E.Label, E.Button
    __ = {('Frame', 'Label', 'Button')[n]: element for n, element in enumerate((Frame._, Label._, Button._))}


class Inherit(Type):
    class Tk(E.Tk):
        ...

    class Style(E.Style):
        ...

    class Toplevel(E.Toplevel):
        ...
    __ = {('TK', 'Style', 'Toplevel')[n]: element._ for n, element in enumerate((Tk, Style, Toplevel))}


class Generate(Type):
    _ = Frame, Label, Button, StringVar, PhotoImage = E.Frame, E.Label, E.Button, E.StringVar, E.PhotoImage
    __: dict[str, list[E.Any]] = {element.__name__: element._ for element in _}

    def __class_getitem__(cls, item: str): return getattr(cls, item, False)


WIDGET = Widget
INHERIT = Inherit
GENERATE = Generate


class TkData:
    S: dict[str, list[E.Any]] = GENERATE.__ | INHERIT.__
    W = {n[0]: WIDGET.__[n] for n in WIDGET.__} | WIDGET.__
    data: dict[str, E.Any]

    class TK_(E.Tk):
        ...

    class WM_(E.Toplevel):
        ...
    TK: TK_
    WM: WM_

    def __class_getitem__(cls, name: str) -> E.Any: return {**cls.S | cls.W | cls.data | cls.__dict__}.get(name, False)

    def get(self, __key: str, __default: E.Any | bool = False): return self.__dict__.get(__key, __default)


if __name__ == '__main__':
    w = TkData.TK_()
    w.mainloop()



from typing import Callable
from . import Win32 as W
from .TK import INHERIT, TkData, E
from .Win32 import c, ct, hWnd, s, wt
_ = [c, ct, s, wt, W]


class Path:
    """Returns the absolute path of a source folder that is in the file path"""

    def __init__(self, source: str = '\\', _=__file__) -> None:
        self.source = f'{_.rsplit(source, 1)[0]}{source}\\' if _.rfind(source) != -1 else _[:_.rfind('\\')] + '\\'

    def __call__(self, add_on: str = '') -> str:
        return self.source + add_on


class Call(TkData, INHERIT):
    settings = {
        'Debug': {'Info': True},
        'WM': {'Call': True, 'visible': True, 'title': True},
        'TK': {'Call': True, 'visible': True, 'mainloop': True},
        'Win32': {'Call': True, 'visible': True, 'mainloop': True}}
    szTitle, szWindowClass = 'WM', 'WM'
    data: dict[str, E.Any] = {
        'WM_GETMINMAXINFO': [(152, 152), (1381, 774)],  # 'WM_GETMINMAXINFO': [(-14, -41), (-15, -15)]
        'TK': {}, 'WM': {'title': szTitle}, 'Win32': {}
    }
    W: dict[str, list[E.Any | E.Label | E.Button | E.Frame] | E.Any]
    ICON = Path('WM', __file__)('images\\resize.ico')
    COLORS = {'bar': ['#1f1f1f', '#353535', '#e81123']}
    WndProc: Callable[..., int]
    hWnd: hWnd

    def __init__(self, name: str = '', *_: E.Any, **__: E.Any):
        if name:
            self.data[name] = self.__class__.__dict__[name]
        super(TkData, self).__init__(*_, **__)


class System(Call):
    Call = Call

    def __init__(self, name: str = '', *_: E.Any, **__: E.Any):
        super().__init__(name, *_, **__)
        print('System Called:', self.__class__.__name__) if self.settings['Debug']['Info'] else ...



import ctypes as ct
from ctypes import wintypes as wt
from typing import Any

import _ctypes  # type: ignore

LRESULT: Any = type('LRESULT', (_ctypes._SimpleCData, ), {'_type_': 'q'})  # type: ignore
WNDPROC = ct.WINFUNCTYPE(LRESULT, wt.HWND, wt.UINT, wt.WPARAM, wt.LPARAM)  # type: ignore



"""Win32API for Python with ctypes"""

from ctypes import WinDLL  # type: ignore

from . import constants, structures
from ._core import LRESULT as LRESULT  # noqa: F401
from ._core import WNDPROC as WNDPROC  # noqa: F401
from ._core import Any, ct, wt

gdi32: Any = WinDLL('gdi32', use_last_error=True)
user32: Any = WinDLL('user32', use_last_error=True)
dwmapi: Any = WinDLL('dwmapi', use_last_error=True)
uxtheme: Any = WinDLL('uxtheme', use_last_error=True)
kernel32: Any = WinDLL('kernel32', use_last_error=True)

s, c = structures, constants

hInst = kernel32.GetModuleHandleW(None)


class hWnd(int):
    ...


user32.DefWindowProcW.argtypes = [wt.HWND, wt.UINT, wt.WPARAM, wt.LPARAM]


# define LOWORD(l)         ((WORD)(((DWORD_PTR)(l)) & 0xffff))
def LOWORD(dwValue: wt.LPARAM): return int(dwValue) & 0xffff


# define HIWORD(l)         ((WORD)((((DWORD_PTR)(l)) >> 16) & 0xffff))
def HIWORD(dwValue: wt.LPARAM): return (int(dwValue) >> 16) & 0xffff


# define GET_X_LPARAM(lp)             ((int)(short)LOWORD(lp))
def GET_X_LPARAM(lp: wt.LPARAM): return ct.c_short(LOWORD(lp)).value


# define GET_Y_LPARAM(lp)             ((int)(short)HIWORD(lp))
def GET_Y_LPARAM(lp: wt.LPARAM): return ct.c_short(HIWORD(lp)).value



import tkinter as tk
from tkinter import ttk
from typing import Any, Literal, Optional
from ..Win32 import hWnd


class Element:
    class __Save:
        _: list[Any]

        def __init__(self, *_: Any, **__: Any):
            super().__init__(*_, **__)
            getattr(self, '_', []).append(self)

    class Tk(__Save, tk.Tk):
        def __init__(self, *_: Any, **__: Any):
            super().__init__(*_, **__)
            self.wm_withdraw()
            super().after(5, self.__update)

        def __update(self):
            self.update_idletasks()
            self.hWnd = hWnd(self.wm_frame(), 16)

    class Style(__Save, ttk.Style):

        def map(self, style: str, query_opt: Any = None, **__: Any) -> dict[Any, Any]:
            return super().map(style, query_opt, **__)  # type: ignore

        def layout(self, style: str, layoutspec: Any = None) -> list[tuple[str, dict[str, Any]]]:
            return super().layout(style, layoutspec)  # type: ignore

        def configure(self, style: str, query_opt: Any = None, **__: Any) -> Any:
            return super().configure(style, query_opt, **__)  # type: ignore

    class Widget(__Save, ttk.Widget):

        def state(self, statespec: Optional[str] = None) -> tuple[str, ...]:
            return super().state(statespec)  # type: ignore

        def grid(self, row: int = 0, column: int = 0, sticky: str = 'nsew', **__: Any):  # type: ignore
            super().grid_configure(row=row, column=column, sticky=sticky, **__)

            def configure(weight: dict[int, tuple[int, int]] = {0: (1, 1)}):
                for i in weight:
                    self.rowconfigure(i, weight=weight[i][0])
                    self.columnconfigure(i, weight=weight[i][1])
                return self
            return configure

        def pack(self, fill: Literal['none', 'x', 'y', 'both'] = 'both', expand: int = 1, **__: Any):  # type: ignore
            super().pack(fill=fill, expand=expand, **__)

            def configure(weight: dict[int, tuple[int, int]] = {0: (1, 1)}):
                for i in weight:
                    self.rowconfigure(i, weight=weight[i][0])
                    self.columnconfigure(i, weight=weight[i][1])
                return self
            return configure

    class Frame(Widget, ttk.Frame):
        ...

    class Label(Widget, ttk.Label):
        ...

    class Button(Widget, ttk.Button):
        ...

    class Toplevel(__Save, tk.Toplevel):
        def __init__(self, *_: Any, **__: Any):
            super().__init__(*_, **__)
            self.wm_withdraw()
            super().after(5, self.__update)

        def __update(self):
            self.update_idletasks()
            self.hWnd = hWnd(self.wm_frame(), 16)

    class StringVar(__Save, tk.StringVar):
        ...

    class PhotoImage(__Save, tk.PhotoImage):
        def __generate(self, image: Optional[tk.PhotoImage]) -> tuple[tk.PhotoImage, Literal['copy'], tk.PhotoImage]:
            return (PhotoImage(), 'copy', self) if image is None else (self, 'copy', image)  # type: ignore

        def cropping(self, x1: int, y1: int, x2: int, y2: int, image: Optional[tk.PhotoImage] = None) -> tk.PhotoImage:
            self.tk.call(*self.__generate(image), '-from', x1, y1, x2, y2, '-to', 0, 0)
            return getattr(self, '_')[-1]


class Tk(Element.Tk):
    _: list[Element.Tk] = []


class Style(Element.Style):
    _: list[Element.Style] = []


class Frame(Element.Frame):
    _: list[Element.Frame] = []


class Label(Element.Label):
    _: list[Element.Label] = []


class Button(Element.Button):
    _: list[Element.Button] = []


class Toplevel(Element.Toplevel):
    _: list[Element.Toplevel] = []


class StringVar(Element.StringVar):
    _: list[Element.StringVar] = []


class PhotoImage(Element.PhotoImage):
    _: list[Element.PhotoImage] = []


Widget = Element.Widget



"""Structures for Win32API"""
from ._core import ct, wt, WNDPROC


class POINT(ct.Structure):
    _fields_ = [
        ('x', wt.LONG),
        ('y', wt.LONG),
    ]


class RECT(ct.Structure):
    _fields_ = [
        ('left', wt.LONG),
        ('top', wt.LONG),
        ('right', wt.LONG),
        ('bottom', wt.LONG),
    ]


class MARGINS(ct.Structure):
    _fields_ = [
        ('cxLeftWidth', wt.INT),
        ('cxRightWidth', wt.INT),
        ('cyTopHeight', wt.INT),
        ('cyBottomHeight', wt.INT),
    ]


class MSG(ct.Structure):
    _fields_ = [("hwnd", wt.HWND),
                ("message", wt.UINT),
                ("wParam", wt.WPARAM),
                ("lParam", wt.LPARAM),
                ("time", wt.DWORD),
                ("pt", POINT)]


class WNDCLASSEXW(ct.Structure):
    _fields_ = [
        ('cbSize', wt.UINT),
        ('style', wt.UINT),
        ('lpfnWndProc', WNDPROC),
        ('cbClsExtra', wt.INT),
        ('cbWndExtra', wt.INT),
        ('hInstance', wt.HINSTANCE),
        ('hIcon', wt.HICON),
        ('hCursor', wt.HICON),
        ('hbrBackground', wt.HBRUSH),
        ('lpszMenuName', wt.LPCWSTR),
        ('lpszClassName', wt.LPCWSTR),
        ('hIconSm', wt.HICON),
    ]


class PAINTSTRUCT(ct.Structure):
    _fields_ = [
        ('hdc', wt.HDC),
        ('fErase', wt.BOOL),
        ('rcPaint', wt.RECT),
        ('fRestore', wt.BOOL),
        ('fIncUpdate', wt.BOOL),
        ('rgbReserved', wt.BYTE * 32),
    ]


class NCCALCSIZE_PARAMS(ct.Structure):
    _fields_ = [
        ('rgrc', RECT * 3),
        ('lppos', POINT),
    ]


class MINMAXINFO(ct.Structure):
    _fields_ = [
        ('ptReserved', POINT),
        ('ptMaxSize', POINT),
        ('ptMaxPosition', POINT),
        ('ptMinTrackSize', POINT),
        ('ptMaxTrackSize', POINT),
    ]


class BITMAP(ct.Structure):
    _fields_ = [
        ('bmType', wt.DWORD),
        ('bmWidth', wt.LONG),
        ('bmHeight', wt.LONG),
        ('bmWidthBytes', wt.DWORD),
        ('bmPlanes', wt.WORD),
        ('bmBitsPixel', wt.WORD),
        ('bmBits', wt.LPVOID),
    ]


class BITMAPFILEHEADER(ct.Structure):
    _fields_ = [
        ('bfType', wt.WORD),
        ('bfSize', wt.DWORD),
        ('bfReserved1', wt.WORD),
        ('bfReserved2', wt.WORD),
        ('bfOffBits', wt.DWORD),
    ]


class WINDOWPLACEMENT(ct.Structure):
    _fields_ = [
        ("length", wt.UINT),
        ("flags", wt.UINT),
        ("showCmd", wt.UINT),
        ("ptMinPosition", wt.POINT),
        ("ptMaxPosition", wt.POINT),
        ("rcNormalPosition", wt.RECT),
        ("rcDevice", wt.RECT),
    ]



from .core import Call, Path, System
from .TK import E


class Style(System, E.Style):
    def __init__(self, ):
        super().__init__()

        settings = {
            'TFrame': {
                'configure': {
                    'background': self.COLORS['bar'][0]
                },
            },
            'TLabel': {
                'configure': {
                    'background': self.COLORS['bar'][0],
                    'font': ('Yu Gothic UI', 14),
                    'foreground': '#ffffff',
                },
                'layout': [
                    ('Label.padding', {'children': [(
                        'Label.label', {'sticky': 'nswe'})], 'sticky': 'nswe'})
                ],
            },
            'TButton': {
                'configure': {
                    'background': self.COLORS['bar'][0],
                    'padding': (17, 10, 17, 11),
                },
                'layout': [('Button.padding', {'children': [('Button.label', {'sticky': 'nswe'})], 'sticky': 'nswe'})],
            },
            'bar.TFrame': {
                'configure': {
                    'background': self.COLORS['bar'][0]
                },
                'layout': [('Frame.padding', {'sticky': 'nswe'})],
            },
            'bar.TLabel': {
                'configure': {
                    'background': self.COLORS['bar'][0],
                    'font': ('Yu Gothic UI', 10),
                    'foreground': '#ffffff',
                    'anchor': 'w',
                    'padding': (10, 0, 10, 5),
                },
                'map': {
                    'foreground': [('!background', '#858585')]
                },
                'layout': [
                    ('Label.padding', {'children': [(
                        'Label.label', {'sticky': 'nswe'})], 'sticky': 'nswe'})
                ],
            },
            'bar.TButton': {
                'configure': {
                    'background': self.COLORS['bar'][0],
                    'padding': (17, 10, 17, 11),
                    'anchor': 'nw'
                },
                'layout': [('Button.padding', {'children': [('Button.label', {'sticky': 'nswe'})], 'sticky': 'nswe'})],
            }
        }
        self.theme_create('WM', parent='default', settings=settings)  # type: ignore
        self.theme_use('WM')
        self.set_styles()
        self.set_images()

    def set_styles(self):
        if Call.data['WM'].get('title', False):
            self.W['L'][0]['style'] = 'title.bar.TLabel'
        self.W['F'][0]['style'] = 'bar.TFrame'
        """ for i, s in enumerate(('min', 'max', 'exit')):
            self.W['B'][i]['style'] = f'{s}.TButton'
        # minimize maximize exit """
        self.W['Button'][0]['style'] = 'minimize.bar.TButton'
        self.W['Button'][1]['style'] = 'maximize.bar.TButton'
        self.W['Button'][2]['style'] = 'exit.bar.TButton'

    def set_images(self):
        image = E.PhotoImage(file=Path('WM', __file__)('images\\0-00-000.png'))
        for i in range(3):
            E.PhotoImage().cropping(i * 12, 0, (i + 1) * 12, 12, image)
            self.configure(self.W['B'][i]['style'], image=self.S['PhotoImage'][-1])
            E.PhotoImage().cropping(i * 12, 12, (i + 1) * 12, 24, image)
            self.map(self.W['B'][i]['style'], background=[
                ('active', self.COLORS['bar'][1] if i != 2 else self.COLORS['bar'][2])])
            if i != 1:
                self.map(self.W['B'][i]['style'], image=[
                    ('!alternate', 'active', self.S['PhotoImage'][-2]),
                    ('!alternate', '!background', self.S['PhotoImage'][-1])])
            else:
                self.map(self.W['B'][i]['style'], image=[
                    ('!alternate', 'active', self.S['PhotoImage'][-2]),
                    ('!alternate', '!background', self.S['PhotoImage'][-1]),
                    ('alternate', 'background', E.PhotoImage().cropping(36, 0, 48, 12, image)),
                    ('alternate', 'active', self.S['PhotoImage'][-1]),
                    ('alternate', '!background', E.PhotoImage().cropping(36, 12, 48, 24, image))])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['WM', 'WM.TK', 'WM.Win32']

package_data = \
{'': ['*'], 'WM': ['images/*']}

setup_kwargs = {
    'name': '574d',
    'version': '0.0.0rc0.dev2',
    'description': 'A Window Manager for Tkinter On Windows 10',
    'long_description': "# WM\nA Window Manager for Tkinter On Windows 10\n\n[Video Test](https://user-images.githubusercontent.com/73524758/200260996-18de1043-4b5f-4f9c-b3ee-6e27d1903594.mp4)\n* [WM/Wiki](https://github.com/XxFULLDLCxX/WM/wiki)\n\n## How to install\n```Bash\npip install 574d\n```\n\n## How to test\n> ```Python\n> from WM import TK\n>\n> ...\n>\n> if __name__ == '__main__':\n>     Tk = TK()\n>     Tk.mainloop()\n> ```\n\nI developed a code pattern based on the Sword Art Online (SAO) for Tkinter\n\n###  For Example:\n\n**`System Call Generate`** `Button` **`Element`** `<Object-ID>` _`Discharge`_ !\n\n_Discharge_ is only a SAO reference.\n\n[`./WM/core.py`](./WM/core.py)\n> ```Python\n> class Call(TkData, INHERIT):  # Object-IDs are here.\n>     ...\n>     szTitle, szWindowClass = 'WM', 'WM'\n>     ...\n>\n> class System(Call):  # class TK(System, Call.Tk): \n>     Call = Call  # System.Call\n>     ...\n> ```\n\n [`./WM/views.py`](./WM/views.py)\n> ```python\n> class Element(E.Widget, E.PhotoImage, System, Call):  # type: ignore\n>     def __new__(cls, name: str = '', *_: E.Any, generic: bool = False, **__: E.Any):\n>         E = GENERATE.__dict__[name](*_, **__)\n>         if generic:\n>             del E._[-1]\n>         return E\n>\n>\n> class Generate(System, Call):\n>     def __init__(self):\n>         super(System, self).__init__()  # info when called\n>         ...\n>         # with Element Constructor\n>         Element('Frame', self.TK).grid(0, 0, 'nsew', padx=1, pady=1)({0: (1, 1), 1: (0, 1), 2: (1, 0)}).grid_remove()\n>         self.W['F'][-1].grid()\n>         # another way to do the same\n>         my_frame = ttk.Frame(self.TK)\n>         my_frame.grid(row=0, column=0, sticky='nsew', padx=1, pady=1)\n>         my_frame.grid_rowconfigure(0, weight=1)\n>         my_frame.grid_rowconfigure(2, weight=1)\n>         my_frame.grid_columnconfigure(0, weight=1)\n>         my_frame.grid_columnconfigure(1, weight=1)\n>         my_frame.grid_remove()\n>         my_frame.grid()\n> ```\n\n\n\n\n",
    'author': 'XxFULLDLCxX',
    'author_email': 'XxFULLDLCxX@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from .core import Call, System, W, c, ct, s, wt, E


@W.WNDPROC
class WndProc:

    @classmethod
    def __info__(cls, hWnd: wt.HWND, uMsg: wt.UINT, wParam: wt.WPARAM, lParam: wt.LPARAM):
        match uMsg:
            case c.WM_CREATE: print('WM_CREATE')
            case c.WM_DESTROY: print('WM_DESTROY')
            case _: ...

    def __new__(cls, *_: E.Any):
        cls.__info__(*_) if Call.settings['Debug']['Info'] else ...
        return cls.__TK__(*_) if Call.WM else cls.__Win32__(*_)

    @classmethod
    def __Win32__(cls, hWnd: wt.HWND, uMsg: wt.UINT, wParam: wt.WPARAM, lParam: wt.LPARAM):
        match uMsg:
            case c.WM_DESTROY: W.user32.PostQuitMessage(0)
            case _: return W.user32.DefWindowProcW(hWnd, uMsg, wParam, lParam)
        return 0

    @classmethod
    def __TK__(cls, hWnd: wt.HWND, uMsg: wt.UINT, wParam: wt.WPARAM, lParam: wt.LPARAM):
        match uMsg:
            case c.WM_CREATE: Call.hWnd = W.hWnd(hWnd)
            case c.WM_MOVE:
                Call.data |= {'WM_MOVE': (W.GET_X_LPARAM(lParam), W.GET_Y_LPARAM(lParam))}
                Call.TK.event_generate('<<WM_MOVE>>', x=W.GET_X_LPARAM(lParam), y=W.GET_Y_LPARAM(lParam))
            case c.WM_SIZE:
                __add = -5 if int(wParam) == c.SIZE_MAXIMIZED else 1
                Call.TK.event_generate('<<WM_SIZE>>', rootx=W.LOWORD(lParam) + 2, rooty=W.HIWORD(lParam) + __add)
                Call.data |= {'WM_SIZE': (W.LOWORD(lParam) + 2, W.HIWORD(lParam) + __add)}
                Call.WM.geometry(f'{W.LOWORD(lParam) - 16}x{W.HIWORD(lParam)}+0+0')
                Call.WM.update()
            case c.WM_NCCALCSIZE if wParam:
                pncc = s.NCCALCSIZE_PARAMS.from_address(int(lParam))
                pncc.rgrc[0].bottom -= (0 if Call.WM.state() == 'zoomed' else 8)
                pncc.rgrc[0].right -= 8
                pncc.rgrc[0].left += 8
                pncc.rgrc[0].top += 0
                return 0
            case c.WM_GETMINMAXINFO:
                pmmi = s.MINMAXINFO.from_address(int(lParam))
                pmmi.ptMinTrackSize.x, pmmi.ptMinTrackSize.y = Call.data['WM_GETMINMAXINFO'][0]
                pmmi.ptMaxTrackSize.x, pmmi.ptMaxTrackSize.y = Call.data['WM_GETMINMAXINFO'][1]
            case c.WM_ACTIVATE:
                margins = s.MARGINS()
                if wParam:  # Make sure if the Window is Activate.
                    margins.cyTopHeight = 1  # Create a Top border
                    Call.TK.event_generate('<<Activate>>')
                else:
                    Call.TK.event_generate('<<Inactivate>>')
                W.dwmapi.DwmExtendFrameIntoClientArea(hWnd, ct.byref(margins))
            case c.WM_DESTROY:
                Call.TK.destroy()
                W.user32.PostQuitMessage(0)
            case _: return W.user32.DefWindowProcW(hWnd, uMsg, wParam, lParam)

        return System.data['WndProc'](hWnd, uMsg, wParam, lParam) if System.data.get('WndProc', False) else 0


class Event(System):
    def __init__(self):
        super().__init__()

        self.W['B'][0]['command'] = self.wm_minsize
        self.W['B'][1]['command'] = self.wm_maxsize
        self.W['B'][2]['command'] = self.wm_destroy

        self.WM.bind('<Configure>', self.wm_size)
        self.WM.bind('<Motion>', self.wm_hittest)
        self.TK.bind('<Button-1>', self.wm_focus)

        self.W['L'][0].bind('<Button-1>', self.wm_move)
        # WM_GETMINMAXINFO
        self.W['L'][0].bind('<Double-Button-1>', self.wm_maxsize)

        # WM_ACTIVATE
        self.TK.bind('<<Activate>>', self.wm_activate)
        self.TK.bind('<<Inactivate>>', self.wm_inactivate)

    def wm_size(self, *_: E.Any):
        _add = (35, 40) if self.WM.state() == 'zoomed' else (34, 33)
        self.W['F'][0].grid(pady=7 if self.WM.state() == 'zoomed' else 0)
        self.TK.wm_geometry(f"{self.data['WM_SIZE'][0]-18}x{abs(self.data['WM_SIZE'][1]-_add[0])}+0+{_add[1]}")
        nn = '' if self.WM.state() == 'zoomed' else '!'
        self.WM.tk.call(self.W['B'][1], 'state', nn + 'alternate')

    def wm_focus(self, *_: E.Any):
        W.user32.SetFocus(Call.TK.hWnd)
        W.user32.SendMessageW(Call.hWnd, c.WM_ACTIVATE, c.WA_ACTIVE, 0)

    def wm_move(self, event: E.Any):
        W.user32.ReleaseCapture()
        W.user32.PostMessageW(Call.hWnd, c.WM_NCLBUTTONDOWN, self.wm_hittest(event), 0)

    def wm_hittest(self, event: E.Any):
        ptMouse, rcWindow = s.POINT(event.x_root, event.y_root), s.RECT()
        W.user32.GetWindowRect(Call.hWnd, ct.byref(rcWindow))
        uCol, uRow, BORDER = 1, 1, 6
        if ptMouse.x - rcWindow.left < BORDER:
            uCol = 0
        elif ptMouse.x - rcWindow.left > rcWindow.right - rcWindow.left - BORDER:
            uCol = 2
        if ptMouse.y - rcWindow.top < BORDER:
            uRow = 0
        elif ptMouse.y - rcWindow.top > rcWindow.bottom - rcWindow.top - BORDER:
            uRow = 2

        cursors = (('size_nw_se', 'size_ns', 'size_ne_sw'),
                   ('size_we', 'arrow', 'size_we'),
                   ('size_ne_sw', 'size_ns', 'size_nw_se'))
        event.widget['cursor'] = cursors[uRow][uCol]
        hittests = ((c.HTTOPLEFT, c.HTTOP, c.HTTOPRIGHT),
                    (c.HTLEFT, c.HTCAPTION, c.HTRIGHT),
                    (c.HTBOTTOMLEFT, c.HTBOTTOM, c.HTBOTTOMRIGHT))
        return hittests[uRow][uCol]

    def wm_maxsize(self, *_: E.Any):
        self.wm_restore() if self.WM.state() == 'zoomed' else W.user32.ShowWindow(Call.hWnd, c.SW_MAXIMIZE)

    def wm_minsize(self, *_: E.Any): W.user32.ShowWindow(Call.hWnd, c.SW_MINIMIZE)

    def wm_restore(self, *_: E.Any): W.user32.ShowWindow(Call.hWnd, c.SW_RESTORE)

    def wm_destroy(self, *_: E.Any): W.user32.PostMessageW(Call.hWnd, c.WM_CLOSE, 0, 0)

    def wm_activate(self, *_: E.Any):
        """Called when the window is activated."""
        [element.tk.call(element, 'state', 'background') for element in self.W['B'] + self.W['L']]
        #: self.WM.update() This produces a bug when the inactive window is activated by pressing the bottom edge.
        self.WM.update_idletasks()  # It's necessary so that there is no delay in the update when clicking on the edges.

    def wm_inactivate(self, *_: E.Any):
        """Called when the window is inactivated."""
        [element.tk.call(element, 'state', '!background') for element in self.W['B'] + self.W['L']]



"""Constanst for Win32API"""
# flake8: noqa
# spaces and more spaces

WM_CREATE         = 0x0001
WM_DESTROY        = 0x0002
WM_MOVE           = 0x0003
WM_SIZE           = 0x0005
WM_ACTIVATE       = 0x0006
WM_SETFOCUS       = 0x0007
WM_KILLFOCUS      = 0x0008
WM_CLOSE          = 0x0010
WM_ERASEBKGND     = 0x0014
WM_SYSCOLORCHANGE = 0x0015
WM_ENABLE         = 0x000A
WM_CHILDACTIVATE  = 0x0022
WM_PAINT          = 0x000F
WM_GETMINMAXINFO  = 0x0024
WM_GETHOTKEY      = 0x0033
WM_SETHOTKEY      = 0x0050
WM_NCCALCSIZE     = 0x0083
WM_NCHITTEST      = 0x0084
WM_NCPAINT        = 0x0085
WM_NCACTIVATE     = 0x0086
WM_NCLBUTTONDOWN  = 0x00A1
WM_NCLBUTTONUP    = 0x00A2
WM_INITDIALOG     = 0x0110
WM_COMMAND        = 0x0111
WM_SIZING         = 0x0214
WM_MOVING         = 0x0216
WM_SYSCOMMAND     = 0x0112
WM_LBUTTONUP      = 0x0202
WM_MOUSELEAVE     = 0x02A3


HTNOWHERE, HTCLIENT, HTCAPTION = (0, 1, 2)
HWND_TOP, HWND_BOTTOM, HWND_TOPMOST, HWND_NOTOPMOST = 0, 1, -1, -2
HTLEFT, HTRIGHT, HTTOP, HTTOPLEFT, HTTOPRIGHT, HTBOTTOM, HTBOTTOMLEFT, HTBOTTOMRIGHT = range(10, 18)

WMSZ_LEFT        = 1
WMSZ_RIGHT       = 2
WMSZ_TOP         = 3
WMSZ_TOPLEFT     = 4
WMSZ_TOPRIGHT    = 5
WMSZ_BOTTOM      = 6
WMSZ_BOTTOMLEFT  = 7
WMSZ_BOTTOMRIGHT = 8

LWA_COLORKEY, LWA_ALPHA = 1, 2
WHITE_BRUSH, BLACK_BRUSH, COLOR_WINDOW = 0, 4, 5
GWL_WNDPROC, GWL_STYLE, GWL_EXSTYLE = -4, -16, -20
IDI_APPLICATION, LR_LOADFROMFILE, LR_DEFAULTSIZE, IMAGE_ICON  = 32512, 0x10, 0x0040, 1
SW_HIDE, SW_NORMAL, SW_MINIMIZE, SW_MAXIMIZE, SW_SHOW, SW_RESTORE, SW_SHOWDEFAULT = 0, 1, 2, 3, 5, 9, 10
EXPAND_BORDER, TOPEXTENDWIDTH, BOTTOMEXTENDWIDTH, LEFTEXTENDWIDTH, RIGHTEXTENDWIDTH = 8, 8, 8, 8, 8
SC_MOVE, SC_MINIMIZE, SC_MAXIMIZE, SC_MOUSEMENU, SC_RESTORE, SC_DEFAULT = 0xF010, 0xF020, 0xF030, 0xF090, 0xF120, 0xF160

WS_OVERLAPPED       = 0x00000000
WS_THICKFRAME       = 0x00040000
WS_MAXIMIZEBOX      = 0x00010000
WS_MINIMIZEBOX      = 0x00020000
WS_SYSMENU          = 0x00080000
WS_DLGFRAME         = 0x00400000
WS_BORDER           = 0x00800000
WS_CAPTION          = 0x00C00000
WS_OVERLAPPEDWINDOW = 0x00CF0000
WS_CLIPCHILDREN     = 0x02000000
WS_CLIPSIBLINGS     = 0x04000000
WS_VISIBLE          = 0x10000000
WS_DISABLED         = 0x08000000
WS_CHILDWINDOW      = 0x40000000
WS_CHILD            = 0x40000000
WS_POPUP            = 0x80000000
WS_POPUPWINDOW      = 0x80880000

WS_EX_TOPMOST             = 0x00000008
WS_EX_TRANSPARENT         = 0x00000020
WS_EX_MDICHILD            = 0x00000040
WS_EX_TOOLWINDOW          = 0x00000080
WS_EX_CLIENTEDGE          = 0x00000200
WS_EX_APPWINDOW           = 0x00040000
WS_EX_LAYERED             = 0x00080000
WS_EX_NOREDIRECTIONBITMAP = 0x00200000
WS_EX_COMPOSITED          = 0x02000000
WS_EX_NOACTIVATE          = 0x08000000
WS_EX_ACCEPTFILES         = 0x00000010

GCL_STYLE = -26
CS_VREDRAW, CS_HREDRAW, CS_DBLCLKS, CS_PARENTDC = 0x0001, 0x0002, 0x0008, 0x0080

SWP_NOSIZE         = 0x0001
SWP_NOMOVE         = 0x0002
SWP_NOZORDER       = 0x0004
SWP_NOREDRAW       = 0x0008
SWP_NOACTIVATE     = 0x0010
SWP_FRAMECHANGED   = 0x0020
SWP_SHOWWINDOW     = 0x0040
SWP_HIDEWINDOW     = 0x0080
SWP_NOCOPYBITS     = 0x0100
SWP_NOREPOSITION   = 0x0200
SWP_NOSENDCHANGING = 0x0400
SWP_DEFERERASE     = 0x2000
SWP_ASYNCWINDOWPOS = 0x4000

CW_USEDEFAULT = 0x80000000
WA_INACTIVE, WA_ACTIVE, WA_CLICKACTIVE = 0, 1, 2
INPUT_MOUSE, INPUT_KEYBOARD, INPUT_HARDWARE = 0, 1, 2
PM_NOREMOVE, PM_REMOVE, PM_NOYIELD = 0x0000, 0x0001, 0x0002
SIZE_RESTORED, SIZE_MINIMIZED, SIZE_MAXIMIZED, SIZE_MAXSHOW, SIZE_MAXHIDE = range(5)
KEYEVENTF_EXTENDEDKEY, KEYEVENTF_KEYUP, KEYEVENTF_UNICODE, KEYEVENTF_SCANCODE = 0x0001, 0x0002, 0x0004, 0x0008
IDC_ARROW, IDC_SIZENWSE, IDC_SIZENESW, IDC_SIZEWE, IDC_SIZENS, IDC_SIZEALL = 32512, 32642, 32643, 32644, 32645, 32646
IDS_APP_TITLE = 103
IDI_GDICAPTURINGANIMAGE = 107
IDC_GDICAPTURINGANIMAGE = 109
IDI_SMALL = 108
MAX_LOADSTRING = 100
MB_OK = 0x00000000
HALFTONE = 4
SM_CXSCREEN = 0
SM_CYSCREEN = 1
SRCCOPY = 0x00CC0020
SRCINVERT = 0x00660046
SRCAND = 0x008800C6
BLACKNESS = 0x4200042
CAPTURBLT = 0x40000000
BI_RGB = 0
GMEM_MOVEABLE = 0x0002
GMEM_ZEROINIT = 0x0040
GHND = GMEM_MOVEABLE | GMEM_ZEROINIT
DIB_RGB_COLORS = 0
GENERIC_WRITE = 0x40000000
CREATE_ALWAYS = 2
FILE_ATTRIBUTE_NORMAL = 0x00000080
RDW_INVALIDATE = 0x0001

IDM_ABOUT = 104
IDD_ABOUTBOX = 103
IDM_EXIT = 105

IDOK = 1
IDCANCEL = 2



import re
from typing import Any, Optional, Callable

from .core import Call, System, W, c, ct, s, wt, hWnd
from .event import Event, WndProc
from .style import Style
from .views import Views

setattr(Call, 'dwStyle', c.WS_OVERLAPPEDWINDOW | c.WS_CLIPCHILDREN | c.WS_CLIPSIBLINGS | c.WS_VISIBLE)
setattr(Call, 'dwExStyle', c.WS_EX_LAYERED | c.WS_EX_APPWINDOW | c.WS_EX_ACCEPTFILES)  # | c.WS_EX_TOPMOST


class WinMain(System):
    """Win32API, an implementation of a Win32 window, to disable the normal title bar but keep resizing."""

    def __init__(self, *_: int, **__: int): super().__init__('hWnd', **__)

    def __new__(cls, w: int = 0, h: int = 0, x: int = 50, y: int = 0):
        W.user32.LoadStringW(W.hInst, 109, cls.szWindowClass, c.MAX_LOADSTRING)
        W.user32.LoadStringW(W.hInst, c.IDS_APP_TITLE, cls.szTitle, c.MAX_LOADSTRING)
        dwExStyle, atom, dwStyle = getattr(cls, 'dwExStyle'), cls.resister_class(), getattr(cls, 'dwStyle')
        cls.hWnd = W.user32.CreateWindowExW(dwExStyle, atom, cls.szTitle, dwStyle, x, y, w, h, 0, 0, W.hInst, 0)
        return super().__new__(cls)

    @classmethod
    def resister_class(cls):
        wcex = s.WNDCLASSEXW()
        wcex.cbSize = ct.sizeof(s.WNDCLASSEXW)
        wcex.style = c.CS_HREDRAW | c.CS_VREDRAW
        wcex.lpfnWndProc = getattr(cls, 'WndProc', WndProc)
        wcex.lpszClassName = cls.szWindowClass
        wcex.hInstance = W.hInst
        wcex.hIcon = W.user32.LoadImageW(W.hInst, cls.ICON, c.IMAGE_ICON, 0, 0, c.LR_DEFAULTSIZE | c.LR_LOADFROMFILE)
        wcex.hCursor = W.user32.LoadCursorW(0, c.IDC_ARROW)
        wcex.hbrBackground = W.gdi32.CreateSolidBrush(0x1F1F1F)
        return W.user32.RegisterClassExW(ct.byref(wcex))

    def mainloop(self):
        """A Win32API message loop"""
        if self.settings['Win32']['visible']:
            W.user32.SetLayeredWindowAttributes(self.hWnd, 255, 255, c.LWA_ALPHA)
        if self.settings['Win32']['mainloop']:
            msg = wt.MSG()
            hAccelTable = W.user32.LoadAcceleratorsW(W.hInst, 109)
            while W.user32.GetMessageW(ct.byref(msg), None, 0, 0):
                if not W.user32.TranslateAcceleratorW(msg.hWnd, hAccelTable, ct.byref(msg)):
                    W.user32.TranslateMessage(ct.byref(msg))
                    W.user32.DispatchMessageW(ct.byref(msg))


class WM(System, Call.WM_):
    hWnd: hWnd

    def __call__(self, *_: Any): return [i() for i in _]

    def __init__(self, *_):
        Call.WM = self.WM = self
        super().__init__()
        self.config(bg='#000001')
        self.attributes('-transparentcolor', '#000001')  # type: ignore

    def state(self, newstate: Optional[str] = None):  # type: ignore
        state = {'iconic': c.SW_MINIMIZE, 'zoomed': c.SW_MAXIMIZE, 'normal': c.SW_NORMAL}
        if newstate in ('iconic', 'zoomed', 'normal'):
            W.user32.ShowWindow(Call.hWnd, state.get(newstate))
        elif newstate is not None:
            raise ValueError(f'{newstate} is not an accepted state.')
        lpwndpl = s.WINDOWPLACEMENT()
        W.user32.GetWindowPlacement(Call.hWnd, ct.byref(lpwndpl))
        return {c.SW_MINIMIZE: 'iconic', c.SW_MAXIMIZE: 'zoomed', c.SW_NORMAL: 'normal'}.get(lpwndpl.showCmd, 'normal')


class Merge(System):
    def __init__(self):
        super().__init__()

        if self.settings['WM']['Call']:
            WM()(Views, Style, Event)

        if self.settings['Win32']['Call']:
            Call.hWnd = WinMain(334, 477).hWnd  # Create a Win32 Window Handler on System. 334x510
            self.WM.after(5, self.__settings) if self.settings['WM']['Call'] else ...

        self.TK.event_generate('<<Activate>>')
        self.TK.geometry('320x470+0+0')

    def __settings(self):
        W.user32.SetWindowLongPtrW(Call['TK'].hWnd, c.GWL_STYLE, c.WS_CHILD | c.WS_VISIBLE)
        W.user32.SetWindowLongPtrW(Call['WM'].hWnd, c.GWL_STYLE, c.WS_CHILD | c.WS_VISIBLE)  # c.WS_BORDER |
        W.user32.SetWindowLongPtrW(Call['WM'].hWnd, c.GWL_EXSTYLE, c.WS_EX_NOREDIRECTIONBITMAP)
        W.user32.SetParent(Call['WM'].hWnd, Call['hWnd'])
        W.user32.SetParent(Call['TK'].hWnd, Call['WM'].hWnd)


class TK(Merge, Call.TK_):
    def __call__(self, *_: Any): return [i() for i in _]

    def __init__(self):
        Call.TK = self.TK = self
        super().__init__()
        self['bg'] = '#000000'
        self.maxsize(1366, 766)

    state = WM.state   # type: ignore

    def overrideredirect(self, boolean: Optional[bool] = None):   # type: ignore
        """Not implemented, Use W.SetWindowLongPtrW or Tkinter pure."""
        return super().overrideredirect(boolean)   # type: ignore

    def maxsize(self, width: Optional[int] = None, height: Optional[int] = None) -> tuple[int, int]:   # type: ignore
        super().maxsize(width, height)   # type: ignore
        if width is not None and height is not None:
            self.data['WM_GETMINMAXINFO'][1] = width + 15, height + 8
            W.user32.PostMessageW(Call.hWnd, W.c.WM_GETMINMAXINFO, 0, 0)
        return self.data['WM_GETMINMAXINFO'][1][0] - 15, self.data['WM_GETMINMAXINFO'][1][1] - 8

    def minsize(self, width: Optional[int] = None, height: Optional[int] = None) -> tuple[int, int]:   # type: ignore
        super().minsize(width - 16, height - 39)   # type: ignore
        if width is not None and height is not None:
            self.data['WM_GETMINMAXINFO'][0] = width + 16, height + 39
            W.user32.PostMessageW(Call.hWnd, W.c.WM_GETMINMAXINFO, 0, 0)
        return self.data['WM_GETMINMAXINFO'][0][0] - 16, self.data['WM_GETMINMAXINFO'][0][1] - 39

    def after(self, ms: int, func: Callable[..., None] = None, *_):   # type: ignore
        print(self, '"after" with WM_MOVE produces an error.') if self.settings['Debug']['Info'] else ...
        return super().after(ms, func, *_)

    def title(self, string: str = ''):   # type: ignore
        self.data['WM']['title'] = string if string else self.data['WM'].get('title', '')
        self.S['StringVar'][0].set(self.data['WM']['title']) if self.S['StringVar'] else ...
        return string

    def geometry(self, new_geometry: str = ''):   # type: ignore
        def merge(n: Any): return tuple(map(lambda n: n[0] if n[0] else n[1], n))
        n = merge(zip(*re.findall(r'^(\d+)x(\d+)|([-+]\d+)([-+]\d+)|.*$', new_geometry if new_geometry else '...')[:2]))
        width, height, x, y = map(int, merge(zip(n, (*self.data['WM_SIZE'], *self.data['WM_MOVE']))))
        W.user32.SetWindowPos(Call.hWnd, 0, x, y, width + 16, height + 41, c.SWP_FRAMECHANGED) if any(n) else ...
        WM_SIZE, WM_MOVE = self.data['WM_SIZE'], self.data['WM_MOVE']
        _add, _h = (8, 3) if self.WM.state() == 'zoomed' else (0, 0)
        return '{}x{}{:+}{:+}'.format(WM_SIZE[0] - 2, WM_SIZE[1] - 34 - _h, WM_MOVE[0] - 8 + _add, WM_MOVE[1] + _add)
    """In pure Tkinter, and window state is 'zoom', window position is to 'normal' state. That doesn't happen here."""

    def winfo_x(self) -> int: return self.data['WM_MOVE'][0] - 8

    def winfo_y(self): return self.data['WM_MOVE'][1]

    def winfo_width(self): return self.data['WM_SIZE'][0] - 2

    def winfo_height(self): return self.data['WM_SIZE'][1] - (35 if self.WM.state() == 'zoomed' else 34)

    def winfo_rootx(self) -> int: return self.data['WM_MOVE'][0]

    def winfo_rooty(self) -> int: return self.data['WM_MOVE'][1] + (40 if self.WM.state() == 'zoomed' else 33)

    def mainloop(self, n: int = 0):
        if self.settings['WM']['visible']:
            W.user32.SetLayeredWindowAttributes(Call.hWnd, 255, 255, c.LWA_ALPHA)
        return super().mainloop(n)



from .core import Call, System
from .TK import GENERATE, E


class Element(E.Widget, E.PhotoImage, System, Call):  # type: ignore
    def __new__(cls, name: str = '', *_: E.Any, generic: bool = False, **__: E.Any):
        E = GENERATE.__dict__[name](*_, **__)
        if generic:
            del E._[-1]
        return E


class Generate(System, Call):
    StringVar = GENERATE.StringVar._

    def __init__(self, ):
        self.WM.columnconfigure(0, weight=1)
        Element('Frame', self.WM).grid(ipady=1)({0: (1, 1)})
        Element('Label', self.W['F'][-1], textvar=Element('StringVar')).grid(pady=(0, 3))
        self.StringVar[0].set(self.data['WM']['title']) if self.data['WM'].get('title', False) else ...
        Element('Button', self.W['F'][-1]).grid(0, 3, pady=1)
        Element('Button', self.W['F'][-1]).grid(0, 4, pady=1)
        Element('Button', self.W['F'][-1]).grid(0, 5, pady=1)


class Views(System, Call):
    def __init__(self):
        super().__init__()
        Generate()
